@@ -35,7 +35,7 @@ pub struct Sound
 
 impl Sound
 {
-    fn from_wav(reader: Box<dyn ReadSeek>) -> Result<Self, DataError>
+    fn from_wav<'a>(reader: Box<dyn ReadSeek + 'a>) -> Result<Self, DataError>
     {
         let mut wav_reader = try_or_else!(hound::WavReader::new(reader),
             |error| { Err(DataError::BadData(format!("{}", error))) });
@@ -75,7 +75,7 @@ impl Sound
         })
     }
 
-    fn from_ogg(reader: Box<dyn ReadSeek>) -> Result<Self, DataError>
+    fn from_ogg<'a>(reader: Box<dyn ReadSeek + 'a>) -> Result<Self, DataError>
     {
         let mut ogg_reader = try_or_else!(lewton::inside_ogg::OggStreamReader::new(reader),
             |error| { Err(DataError::BadData(format!("{}", error))) });
@@ -119,13 +119,15 @@ impl DataObject for Sound
     {
         pathname.ends_with(".wav") || pathname.ends_with(".ogg")
     }
-    fn from_io(reader: Box<dyn ReadSeek>, full_pathname : &str, _source : &mut Box<dyn Source>) -> Result<Self, DataError> where Self : Sized
+    fn from_package_source(source : &mut Box<dyn Source>, package_name : &str, pathname : &str) -> Result<Self, DataError> where Self : Sized
     {
-        if full_pathname.ends_with(".wav")
+        let reader = source.read_file(package_name, pathname)?;
+
+        if pathname.ends_with(".wav")
         {
             return Sound::from_wav(reader);
         }
-        else if full_pathname.ends_with(".ogg")
+        else if pathname.ends_with(".ogg")
         {
             return Sound::from_ogg(reader);
         }
