@@ -1,7 +1,7 @@
 //! Gameloop and [gameloop::GameControl] context, which is used to control the game globally
 
 use crate::input::Input;
-use crate::scene::{Scene, BaseScene, SceneType, SceneConfig};
+use crate::scene::{Scene, BaseScene, SceneType, SceneConfig, NullSpawnable};
 use crate::audio::{Audio, AlAudio, NullAudio, AudioError};
 use crate::renderer::{Renderer, RendererError, RendererNewInfo, ViewportMode};
 #[cfg(feature = "imgui_base")] use crate::imgui_handler::ImGuiHandler;
@@ -26,6 +26,8 @@ use std::cell::RefCell;
 use std::collections::HashMap;
 use std::any::TypeId;
 use downcast_rs::Downcast;
+use std::sync::{Arc, Mutex, MutexGuard};
+use crate::doglog::ConsoleHistory;
 
 const MILLISECONDS_PER_NANOSECOND : f64 = 0.000001;
 const LOW_ENERGY_TIMER_MIN : u32 = 60;
@@ -60,7 +62,7 @@ struct PlaceholderScene
 
 impl SceneType for PlaceholderScene
 {
-    type SpawnableIdType = u32;
+    type SpawnableIdType = NullSpawnable;
     
     fn new() -> Self where Self : Sized { PlaceholderScene {} }
     fn config(&mut self, _game: &mut GameControl) -> SceneConfig<Self>
@@ -280,6 +282,11 @@ pub struct GameInfo
     ///  backend cannot be initialized, the second one is attempted, and etc. Recommend keeping
     ///  this at the default unless you know what you're doing.
     pub backends : String,
+    /// Whether to print basic stats (number of entities and components, framerate, basic timing information)
+    ///  to the log. Does nothing on release builds.
+    pub log_stats : bool,
+    /// A list of additional packages to include in the filter used by the default logger (DogLog).
+    pub log_packages : Vec<&'static str>,
     /// Enable kiosk mode, which forces fullscreen and disables closing the window.
     ///  Note that the user may still be able to close the program depending on your
     ///  operating environment's global shortcuts.
@@ -308,6 +315,8 @@ impl Default for GameInfo
             frame_interpolation : FrameInterpolationMode::Off,
             frame_interpolation_cap : None,
             backends : crate::renderer::available_backends(),
+            log_stats : true,
+            log_packages : Vec::new()
         }
     }
 }
@@ -410,6 +419,11 @@ impl GameloopStopwatch
     {
         self.frames_this_second
     }
+
+    pub fn current_ms(&self) -> f64
+    {
+        self.jiffies_this_frame.as_nanos() as f64 * MILLISECONDS_PER_NANOSECOND
+    }
     
     pub fn average_ms(&self) -> f64
     {
@@ -424,6 +438,24 @@ pub trait GameSingleton : Downcast
 
 impl_downcast!(GameSingleton);
 
+pub struct GameControlOptions<'a>
+{
+    console_history: Arc<Mutex<ConsoleHistory>>,
+    sdl_context : Option<&'a sdl2::Sdl>
+}
+
+impl<'a> GameControlOptions<'a>
+{
+    pub fn headless() -> GameControlOptions<'a>
+    {
+        GameControlOptions
+        {
+            console_history : Arc::new(Mutex::new(ConsoleHistory::new())),
+            sdl_context : None
+        }
+    }
+}
+
 /// Main context. You don't normally need to create this yourself as the [Gameloop]
 ///  has its own instance. It can be useful to create this directly if you need to run
 ///  automated tests.
@@ -445,6 +477,8 @@ pub struct GameControl
     source_manager : Rc<RefCell<SourceManager>>,
     resources : DataMultistore,
     singletons : HashMap<TypeId, Box<dyn GameSingleton>>,
+    console_history : Arc<Mutex<ConsoleHistory>>,
+    stats : GameStats,
     focused : bool,
     done : bool,
     #[cfg(feature = "imgui_base")]
@@ -454,7 +488,7 @@ pub struct GameControl
 impl GameControl
 {
     /// (ADVANCED) Creates the main game context
-    pub fn new(gameinfo : &GameInfo, sdl_context : Option<&sdl2::Sdl>) -> Result<GameControl, InitError>
+    pub fn new(gameinfo : &GameInfo, options : GameControlOptions) -> Result<GameControl, InitError>
     {   
         let source_manager = Rc::new(RefCell::new(SourceManager::new()));
         
@@ -467,7 +501,7 @@ impl GameControl
         let controller_subsystem;
         let mut input;
 
-        if let Some(sdl) = sdl_context
+        if let Some(sdl) = options.sdl_context
         {
             let video_subsystem = try_or_else!(sdl.video(),
                 |error| Err(InitError::Sdl2InitError(Sdl2Subsystem::Video, error)));
@@ -522,6 +556,8 @@ impl GameControl
         #[cfg(feature = "imgui_base")]
         try_or_else!(renderer.control().init_imgui_renderer(imgui_handler.imgui_mut()),
             |error| Err(InitError::RendererInitError(error)));
+
+        input.update_mouse_metrics(&renderer);
         
         Ok(GameControl
         {
@@ -539,6 +575,8 @@ impl GameControl
             next_scene_waiting : false,
             source_manager,
             resources,
+            console_history : options.console_history,
+            stats : GameStats::new(),
             focused : true,
             done : false,
             singletons : HashMap::new(),
@@ -706,6 +744,16 @@ impl GameControl
         None
     }
 
+    pub fn console_history(&self) -> MutexGuard<'_, ConsoleHistory>
+    {
+        self.console_history.lock().unwrap()
+    }
+
+    pub fn stats(&self) -> &GameStats
+    {
+        &self.stats
+    }
+
     pub fn goto_new_scene<T : SceneType + 'static>(&mut self)
     {
         let next_scene = Box::new(Scene::<T>::new());
@@ -735,6 +783,7 @@ struct GameloopTimeState
     target_frametime : Duration,
     cap_frametime : Duration,
     last_display : Instant,
+    lag_this_frame : u32,
     lag_this_second : u32,
     thought_yet : bool,
     think_timer : GameloopStopwatch,
@@ -764,6 +813,7 @@ impl GameloopTimeState
             target_frametime,
             cap_frametime,
             last_display,
+            lag_this_frame : 0,
             lag_this_second : 0,
             thought_yet : false,
             think_timer : GameloopStopwatch::new(),
@@ -834,6 +884,59 @@ impl GameloopTimeState
     }
 }
 
+#[derive(Clone)]
+pub struct GameStats
+{
+    pub draw_fps : u32,
+    pub think_fps : u32,
+    pub frame_ms_think : f64,
+    pub frame_ms_draw : f64,
+    pub frame_ms_present : f64,
+    pub frame_ms_wait : f64,
+    pub average_ms_think : f64,
+    pub average_ms_draw : f64,
+    pub average_ms_present : f64,
+    pub average_ms_wait : f64
+}
+
+impl GameStats
+{
+    pub fn new() -> GameStats
+    {
+        GameStats
+        {
+            draw_fps : 0,
+            think_fps : 0,
+            frame_ms_think : 0.0,
+            frame_ms_draw : 0.0,
+            frame_ms_present : 0.0,
+            frame_ms_wait : 0.0,
+            average_ms_think : 0.0,
+            average_ms_draw : 0.0,
+            average_ms_present : 0.0,
+            average_ms_wait : 0.0
+        }
+    }
+
+    fn update_frame_timings(&mut self, time : &GameloopTimeState)
+    {
+        self.frame_ms_think = time.think_timer.current_ms();
+        self.frame_ms_draw = time.draw_timer.current_ms();
+        self.frame_ms_present = time.present_timer.current_ms();
+        self.frame_ms_wait = time.wait_timer.current_ms();
+    }
+
+    fn update_second_timings(&mut self, time : &GameloopTimeState)
+    {
+        self.draw_fps = time.draw_timer.frames_this_second();
+        self.think_fps = time.think_timer.tics_this_second();
+        self.average_ms_think = time.think_timer.average_ms();
+        self.average_ms_draw = time.draw_timer.average_ms();
+        self.average_ms_present = time.present_timer.average_ms();
+        self.average_ms_wait = time.wait_timer.average_ms();
+    }
+}
+
 /// Holds the main game context and performs event handling and timekeeping
 ///  automatically. One of the first structs you will instantiate in your program.
 pub struct Gameloop
@@ -842,9 +945,10 @@ pub struct Gameloop
     sdl_context : sdl2::Sdl,
     event_pump : sdl2::EventPump,
     event_queue : Vec<sdl2::event::Event>,
-    control : GameControl,
+    control : Box<GameControl>,
     current_scene : Box<dyn BaseScene>,
-    time : GameloopTimeState,
+    time : Box<GameloopTimeState>,
+    log_stats : bool,
     kiosk_mode : bool
 }
 
@@ -863,18 +967,25 @@ impl Gameloop
         {
             panic!("Package name cannot be empty.");
         }
+
+        let console_history = Arc::new(Mutex::new(ConsoleHistory::new()));
         
         #[cfg(feature = "default_logger")]
         {
-            let mut logger = doglog::DogLog::new();
+            let mut logger = doglog::DogLog::new(console_history.clone());
             logger.add_package_filter(String::from(gameinfo.package_name));
 
+            for package in &gameinfo.log_packages
+            {
+                logger.add_package_filter(package.to_string());
+            }
+
             #[cfg(debug_assertions)]
             let filter = log::LevelFilter::Debug;
             #[cfg(not(debug_assertions))]
             let filter = log::LevelFilter::Info;
             
-            if let Err(_) = doglog::DogLog::init(logger, filter)
+            if let Err(_) = logger.install(filter)
             {
                 eprintln!("Could not initialize default logger.");
             }
@@ -886,7 +997,7 @@ impl Gameloop
         
         apply_config(&mut gameinfo);
         
-        print_log_section("🐶 KEESHOND Game Engine 🐶");
+        print_log_section("KEESHOND Game Engine");
         
         let version = sdl2::version::version();
         info!("Starting SDL version {}", version);
@@ -898,7 +1009,11 @@ impl Gameloop
             |error| Err(InitError::Sdl2InitError(Sdl2Subsystem::Event, error)));
         
         #[allow(unused_mut)] // mut used only by imgui feature
-        let mut control = try_or_else!(GameControl::new(&gameinfo, Some(&sdl_context)), |error| Err(error));
+        let mut control = Box::new(try_or_else!(GameControl::new(&gameinfo, GameControlOptions
+        {
+            console_history,
+            sdl_context : Some(&sdl_context)
+        }), |error| Err(error)));
         
         control.renderer_mut().set_base_size(gameinfo.base_width as f32, gameinfo.base_height as f32);
         control.renderer_mut().set_window_title(&gameinfo.friendly_name);
@@ -912,7 +1027,8 @@ impl Gameloop
             event_queue : Vec::new(),
             control,
             current_scene : Box::new(Scene::<PlaceholderScene>::new()),
-            time : GameloopTimeState::new(),
+            time : Box::new(GameloopTimeState::new()),
+            log_stats : gameinfo.log_stats,
             kiosk_mode : gameinfo.kiosk_mode
         })
     }
@@ -929,14 +1045,21 @@ impl Gameloop
 
     fn think(&mut self)
     {
-        self.time.think_timer.begin_frame_timing();
-        
         let target_ms = 1000.0 / self.control.target_framerate();
         let cap_ms = 1000.0 / self.control.frame_interpolation_cap().unwrap_or(1000.0);
         self.time.target_frametime = Duration::from_nanos((target_ms * 1_000_000.0) as u64);
         self.time.cap_frametime = Duration::from_nanos((cap_ms * 1_000_000.0) as u64);
+
+        if self.time.jiffy_queue <= self.time.target_frametime
+        {
+            return;
+        }
+
+        self.time.think_timer.begin_frame_timing();
         
         self.control.sync_audio_store();
+
+        self.time.lag_this_frame = 0;
         
         while self.time.jiffy_queue > self.time.target_frametime
         {
@@ -946,6 +1069,7 @@ impl Gameloop
                 while self.time.jiffy_queue > self.time.target_frametime
                 {
                     self.time.jiffy_queue -= self.time.target_frametime;
+                    self.time.lag_this_frame += 1;
                     self.time.lag_this_second += 1;
                 }
                 break;
@@ -995,7 +1119,7 @@ impl Gameloop
         {
             let mut interpolation_amount = 0.0;
 
-            if self.control.frame_interpolation_active() && !self.low_energy_timer_expired()
+            if self.control.frame_interpolation_active() && self.time.lag_this_frame == 0 && !self.low_energy_timer_expired()
             {
                 interpolation_amount = (self.time.jiffy_queue.as_nanos() as f64
                     / self.time.target_frametime.as_nanos() as f64) - 1.0;
@@ -1122,6 +1246,7 @@ impl Gameloop
                 WindowEvent::SizeChanged { .. } =>
                 {
                     self.control.renderer.recalculate_viewport();
+                    self.control.input.update_mouse_metrics(&self.control.renderer);
                 }
                 _ => ()
             },
@@ -1133,7 +1258,8 @@ impl Gameloop
                     && !self.kiosk_mode
                 {
                     self.control.renderer.toggle_fullscreen();
-                } else if !repeat
+                }
+                else if !repeat
                 {
                     if let Some(some_scancode) = scancode
                     {
@@ -1148,6 +1274,22 @@ impl Gameloop
                     self.control.input.update_key(some_scancode.name(), false, true);
                 }
             },
+            Event::MouseMotion { x, y, xrel, yrel, .. } =>
+            {
+                self.control.input.update_mouse_movement(x, y, xrel, yrel);
+            },
+            Event::MouseButtonDown { mouse_btn, .. } =>
+            {
+                self.control.input.update_mouse_button(mouse_btn, true);
+            },
+            Event::MouseButtonUp { mouse_btn, .. } =>
+            {
+                self.control.input.update_mouse_button(mouse_btn, false);
+            },
+            Event::MouseWheel { x, y, .. } =>
+            {
+                self.control.input.update_mouse_wheel(x, y);
+            },
             Event::ControllerDeviceAdded { which, .. } =>
             {
                 if let Some(controller_subsystem) = &self.control.controller_subsystem
@@ -1155,14 +1297,14 @@ impl Gameloop
                     match controller_subsystem.open(which)
                     {
                         Ok(controller) =>
-                            {
-                                info!("Attached controller {}: {}", which, controller.name());
-                                self.control.controllers.insert(which, controller);
-                            },
+                        {
+                            info!("Attached controller {}: {}", which, controller.name());
+                            self.control.controllers.insert(which, controller);
+                        },
                         Err(error) =>
-                            {
-                                error!("Could not open controller: {}", error.to_string());
-                            }
+                        {
+                            error!("Could not open controller: {}", error.to_string());
+                        }
                     }
                 }
             }
@@ -1193,28 +1335,36 @@ impl Gameloop
     
     fn update_fps(&mut self)
     {
+        self.control.stats.update_frame_timings(&self.time);
+
         if self.time.jiffies < self.time.last_fps_jiffies + Duration::from_secs(1)
         {
             return;
         }
 
+        self.control.stats.update_second_timings(&self.time);
+
         let lerp_active = match self.control.frame_interpolation_active()
         {
             true => { "ON" }
             false => { "OFF" }
         };
-        
-        debug!("{} ent {} com  |  {} / {} FPS  |  avg ms  think {:.2}  draw {:.2}  present {:.2}  wait {:.2}  |  lerp {}",
-            self.current_scene.entity_count(),
-            self.current_scene.component_count(),
-            self.time.draw_timer.frames_this_second(),
-            self.time.think_timer.tics_this_second(),
-            self.time.think_timer.average_ms(),
-            self.time.draw_timer.average_ms(),
-            self.time.present_timer.average_ms(),
-            self.time.wait_timer.average_ms(),
-            lerp_active);
-        
+
+        if self.log_stats
+        {
+            debug!("{} ent {} com  |  {} / {} FPS  |  avg ms  think {:.2}  draw {:.2}  present {:.2}  wait {:.2}  |  lerp {}",
+                   self.current_scene.entity_count(),
+                   self.current_scene.component_count(),
+                   self.time.draw_timer.frames_this_second(),
+                   self.time.think_timer.tics_this_second(),
+                   self.time.think_timer.average_ms(),
+                   self.time.draw_timer.average_ms(),
+                   self.time.present_timer.average_ms(),
+                   self.time.wait_timer.average_ms(),
+                   lerp_active);
+        }
+
+        #[cfg(debug_assertions)]
         if self.time.lag_this_second > 0
         {
             warn!("Too much CPU work, game is slowing down!");
@@ -1223,63 +1373,92 @@ impl Gameloop
         self.time.new_second();
     }
 
-    pub fn run(&mut self)
+    pub fn single_frame(&mut self, do_sleep : bool) -> bool
+    {
+        if self.low_energy_timer_expired() && !self.control.focused && do_sleep
+        {
+            self.event_queue.push(self.event_pump.wait_event());
+
+            self.time.flush();
+        }
+
+        self.control.input.pre_update();
+
+        // Consume events
+        for event in self.event_pump.poll_iter()
+        {
+            self.event_queue.push(event);
+        }
+
+        let mut event_queue = Vec::new();
+
+        std::mem::swap(&mut self.event_queue, &mut event_queue);
+
+        for event in event_queue.drain(..)
+        {
+            if !self.process_event(event)
+            {
+                return false;
+            }
+        }
+
+        std::mem::swap(&mut self.event_queue, &mut event_queue);
+
+        self.think();
+        self.draw();
+        if do_sleep
+        {
+            self.wait();
+        }
+        else
+        {
+            self.time.update_jiffies();
+        }
+        self.present();
+
+        self.update_fps();
+
+        // Go to next scene if one is requested
+        if self.control.next_scene_waiting
+        {
+            self.current_scene.end(&mut self.control);
+
+            print_log_section("NEW SCENE");
+
+            std::mem::swap(&mut self.current_scene, &mut self.control.next_scene);
+            self.control.next_scene = Box::new(Scene::<PlaceholderScene>::new());
+
+            self.current_scene.start(&mut self.control);
+
+            // Reset jiffy queue
+            self.time.flush();
+            self.control.next_scene_waiting = false;
+        }
+
+        return !self.control.done;
+    }
+
+    #[allow(unused_mut)]
+    pub fn run(mut self)
+    {
+        #[cfg(not(target_os = "emscripten"))]
+        self.run_desktop();
+        #[cfg(target_os = "emscripten")]
+        self.run_web();
+    }
+
+    #[cfg(not(target_os = "emscripten"))]
+    pub fn run_desktop(&mut self)
     {
         print_log_section("START GAMELOOP");
         
-        self.time = GameloopTimeState::new();
-        
-        'runloop: while !self.control.done
+        self.time = Box::new(GameloopTimeState::new());
+
+        loop
         {
-            if self.low_energy_timer_expired() && !self.control.focused
+            if !self.single_frame(true)
             {
-                self.event_queue.push(self.event_pump.wait_event());
-
-                self.time.flush();
-            }
-
-            // Consume events
-            for event in self.event_pump.poll_iter()
-            {
-                self.event_queue.push(event);
-            }
-
-            let mut event_queue = Vec::new();
-
-            std::mem::swap(&mut self.event_queue, &mut event_queue);
-
-            for event in event_queue.drain(..)
-            {
-                if !self.process_event(event)
-                {
-                    break 'runloop;
-                }
-            }
-
-            std::mem::swap(&mut self.event_queue, &mut event_queue);
-            
-            self.think();
-            self.draw();
-            self.wait();
-            self.present();
-            
-            self.update_fps();
-            
-            // Go to next scene if one is requested
-            if self.control.next_scene_waiting
-            {
-                self.current_scene.end(&mut self.control);
-                
-                print_log_section("NEW SCENE");
-                
-                std::mem::swap(&mut self.current_scene, &mut self.control.next_scene);
-                self.control.next_scene = Box::new(Scene::<PlaceholderScene>::new());
-
-                self.current_scene.start(&mut self.control);
-                
-                // Reset jiffy queue
-                self.time.flush();
-                self.control.next_scene_waiting = false;
+                break;
             }
         }
         
@@ -1287,4 +1466,30 @@ impl Gameloop
         
         self.current_scene.end(&mut self.control);
     }
+
+    #[cfg(target_os = "emscripten")]
+    pub fn run_web(mut self)
+    {
+        print_log_section("START GAMELOOP");
+
+        self.time = Box::new(GameloopTimeState::new());
+
+        let loop_closure = move ||
+        {
+            match self.single_frame(false)
+            {
+                true =>
+                {
+                    crate::thirdparty::emscripten::MainLoopEvent::Continue
+                },
+                false =>
+                {
+                    self.current_scene.end(&mut self.control);
+                    return crate::thirdparty::emscripten::MainLoopEvent::Terminate;
+                }
+            }
+        };
+
+        crate::thirdparty::emscripten::set_main_loop_callback(loop_closure);
+    }
 }
