@@ -3,6 +3,9 @@
 use std::collections::{HashMap, HashSet};
 
 use sdl2::controller::{Button, Axis};
+use sdl2::mouse::MouseButton;
+
+use crate::renderer::Renderer;
 
 const ANGLE_UP_RIGHT_RIGHT : f64 = -std::f64::consts::FRAC_PI_8;
 const ANGLE_UP_RIGHT : f64 = -std::f64::consts::FRAC_PI_4;
@@ -101,13 +104,6 @@ enum CompositeAxisData
     }
 }
 
-#[derive(Debug, Clone)]
-struct DeadzoneData
-{
-    composite_axis : CompositeAxisData,
-    deadzone : f64
-}
-
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum InputDirection4
 {
@@ -145,14 +141,22 @@ fn apply_deadzone(deadzone_opt : Option<&f64>, value : f64) -> f64
 pub struct Input
 {
     keys_held : HashSet<String>,
+    keys_frame_pressed : HashSet<String>,
+    keys_same_frame_released : HashSet<String>,
     axis_values : HashMap<String, AxisData>,
     actions_to_keys : HashMap<String, Vec<String>>,
     action_map : HashMap<String, InputAction>,
+    mouse_to_key : HashMap<MouseButton, String>,
     gamepad_to_key : HashMap<Button, String>,
     gamepad_to_axis_pos : HashMap<Axis, String>,
     gamepad_to_axis_neg : HashMap<Axis, String>,
     composite_axes : HashMap<String, CompositeAxisData>,
     gamepad_deadzones : HashMap<String, f64>,
+    cursor_pos : (f64, f64),
+    cursor_pos_raw : (f64, f64),
+    cursor_motion : (f64, f64),
+    mouse_offset : (f64, f64),
+    mouse_scale : (f64, f64),
     mouse_util : Option<sdl2::mouse::MouseUtil>
 }
 
@@ -160,6 +164,14 @@ impl Input
 {
     pub(crate) fn new(mouse_util : Option<sdl2::mouse::MouseUtil>) -> Input
     {
+        let mut mouse_to_key = HashMap::new();
+
+        mouse_to_key.insert(MouseButton::Left, String::from("mouse button left"));
+        mouse_to_key.insert(MouseButton::Middle, String::from("mouse button middle"));
+        mouse_to_key.insert(MouseButton::Right, String::from("mouse button right"));
+        mouse_to_key.insert(MouseButton::X1, String::from("mouse button back"));
+        mouse_to_key.insert(MouseButton::X2, String::from("mouse button forward"));
+
         let mut gamepad_to_key = HashMap::new();
 
         gamepad_to_key.insert(Button::A, String::from("gamepad a"));
@@ -223,14 +235,22 @@ impl Input
         Input
         {
             keys_held : HashSet::new(),
+            keys_frame_pressed : HashSet::new(),
+            keys_same_frame_released : HashSet::new(),
             axis_values: HashMap::new(),
             actions_to_keys : HashMap::new(),
             action_map : HashMap::new(),
+            mouse_to_key,
             gamepad_to_key,
             gamepad_to_axis_pos,
             gamepad_to_axis_neg,
             composite_axes,
             gamepad_deadzones,
+            cursor_pos : (0.0, 0.0),
+            cursor_pos_raw : (0.0, 0.0),
+            cursor_motion : (0.0, 0.0),
+            mouse_offset : (0.0, 0.0),
+            mouse_scale : (0.0, 0.0),
             mouse_util
         }
     }
@@ -446,6 +466,16 @@ impl Input
         return None;
     }
 
+    pub fn cursor_position(&self) -> (f64, f64)
+    {
+        self.cursor_pos
+    }
+
+    pub fn cursor_motion(&self) -> (f64, f64)
+    {
+        self.cursor_motion
+    }
+
     pub fn cursor_visible(&self) -> bool
     {
         if let Some(mouse_util) = &self.mouse_util
@@ -463,6 +493,24 @@ impl Input
             mouse_util.show_cursor(visible);
         }
     }
+
+    pub fn cursor_locked(&self) -> bool
+    {
+        if let Some(mouse_util) = &self.mouse_util
+        {
+            return mouse_util.relative_mouse_mode();
+        }
+
+        false
+    }
+
+    pub fn set_cursor_locked(&mut self, locked : bool)
+    {
+        if let Some(mouse_util) = &self.mouse_util
+        {
+            mouse_util.set_relative_mouse_mode(locked);
+        }
+    }
     
     pub(crate) fn update_key(&mut self, key : &str, pressed : bool, update_axis : bool)
     {
@@ -471,6 +519,7 @@ impl Input
         if pressed
         {
             self.keys_held.insert(lower_key.clone());
+            self.keys_frame_pressed.insert(lower_key.clone());
 
             if update_axis
             {
@@ -479,7 +528,16 @@ impl Input
         }
         else
         {
-            self.keys_held.remove(&lower_key);
+            // If pressed and released on same frame, treat as pressed this frame and then
+            //  released next frame.
+            if self.keys_frame_pressed.contains(&lower_key)
+            {
+                self.keys_same_frame_released.insert(lower_key.clone());
+            }
+            else
+            {
+                self.keys_held.remove(&lower_key);
+            }
 
             if update_axis
             {
@@ -488,6 +546,12 @@ impl Input
         }
     }
 
+    pub(crate) fn update_same_frame_key(&mut self, key : &str)
+    {
+        self.update_key(key, true, true);
+        self.update_key(key, false, true);
+    }
+
     pub(crate) fn update_axis(&mut self, axis : &str, value : f64)
     {
         let lower_axis = axis.to_lowercase();
@@ -495,9 +559,53 @@ impl Input
         self.axis_values.insert(lower_axis, AxisData { value, raw : value } );
     }
 
-    pub(crate) fn update_gamepad_key(&mut self, key : Button, pressed : bool)
+    pub(crate) fn update_mouse_movement(&mut self, x : i32, y : i32, x_rel : i32, y_rel : i32)
     {
-        let gamepad_key_string = self.gamepad_to_key[&key].clone();
+        self.cursor_pos_raw = (x as f64, y as f64);
+        self.cursor_motion = (x_rel as f64, y_rel as f64);
+
+        self.scale_mouse_position();
+    }
+
+    fn scale_mouse_position(&mut self)
+    {
+        let (x, y) = self.cursor_pos_raw;
+        let (offset_x, offset_y) = self.mouse_offset;
+        let (scale_x, scale_y) = self.mouse_scale;
+
+        self.cursor_pos = ((x + offset_x) * scale_x, (y + offset_y) * scale_y);
+    }
+
+    pub(crate) fn update_mouse_metrics(&mut self, renderer : &Renderer)
+    {
+        let (base_w, base_h) = renderer.base_size();
+        let rect = renderer.window_scissor();
+
+        self.mouse_offset = (-(rect.x as f64), -(rect.y as f64));
+        self.mouse_scale = (base_w as f64 / rect.w as f64, base_h as f64 / rect.h as f64);
+
+        self.scale_mouse_position();
+    }
+
+    pub(crate) fn update_mouse_button(&mut self, button : MouseButton, pressed : bool)
+    {
+        let mouse_key_string = self.mouse_to_key[&button].clone();
+
+        self.update_key(&mouse_key_string, pressed, true)
+    }
+
+    pub(crate) fn update_mouse_wheel(&mut self, x : i32, y : i32)
+    {
+        if x < 0 { self.update_same_frame_key("mouse wheel left"); }
+        else if x > 0 { self.update_same_frame_key("mouse wheel right"); }
+
+        if y < 0 { self.update_same_frame_key("mouse wheel down"); }
+        else if y > 0 { self.update_same_frame_key("mouse wheel up"); }
+    }
+
+    pub(crate) fn update_gamepad_key(&mut self, button : Button, pressed : bool)
+    {
+        let gamepad_key_string = self.gamepad_to_key[&button].clone();
 
         self.update_key(&gamepad_key_string, pressed, true)
     }
@@ -563,6 +671,11 @@ impl Input
         }
     }
 
+    pub(crate) fn pre_update(&mut self)
+    {
+        self.cursor_motion = (0.0, 0.0);
+    }
+
     pub(crate) fn update_actions(&mut self)
     {
         self.adjust_axes_for_deadzone();
@@ -589,5 +702,15 @@ impl Input
 
             action.update_state(is_held, axis_cumulative);
         }
+
+        // If pressed and released on same frame, treat as pressed this frame and then
+        //  released next frame.
+        for key in &self.keys_same_frame_released
+        {
+            self.keys_held.remove(key);
+        }
+
+        self.keys_frame_pressed.clear();
+        self.keys_same_frame_released.clear();
     }
 }
