@@ -5,26 +5,27 @@
 
 use cgmath::{SquareMatrix, Matrix3};
 
-use std::io::{Read, Cursor};
+use std::io::Read;
 use std::rc::Rc;
 use std::cell::RefCell;
 
 use serde::{Serialize, Deserialize};
 use itertools::Itertools;
 
-use keeshond_datapack::{ReadSeek, DataId, DataObject, DataStore, DataError, PreparedStoreError, DataMultistore, TrustPolicy};
+use keeshond_datapack::{DataId, DataObject, DataStore, DataError, PreparedStoreError, DataMultistore, TrustPolicy};
 use keeshond_datapack::source::Source;
 
 #[cfg(feature = "imgui_base")] use imgui;
 
-use sdl2::surface::Surface;
-use sdl2::pixels::PixelFormatEnum;
-
 use crate::scene::BaseScene;
 use crate::util::{NamedItemStore, Rect, BakedRect};
 #[cfg(feature = "opengl")] use crate::renderer::opengl::GlDrawControl;
 #[cfg(feature = "vulkan")] use crate::renderer::vulkan::VulkanDrawControl;
 
+#[cfg(not(windows))] use std::io::Cursor;
+#[cfg(not(windows))] use sdl2::surface::Surface;
+#[cfg(not(windows))] use sdl2::pixels::PixelFormatEnum;
+
 pub(crate) fn png_to_raw<R : Read>(png_reader : R) -> Result<(Vec<u8>, u32, u32), RendererError>
 {
     let decoder = png::Decoder::new(png_reader);
@@ -104,6 +105,16 @@ impl DrawTransform
     }
 }
 
+pub struct WindowViewportInfo
+{
+    pub scale_offset_x : f32,
+    pub scale_offset_y : f32,
+    pub scale_width : f32,
+    pub scale_height : f32,
+    pub base_multiplier_width : f32,
+    pub base_multiplier_height : f32,
+}
+
 pub(crate) struct ViewportInfo
 {
     pub scale_offset_x : f32,
@@ -305,7 +316,7 @@ pub struct RendererNewInfo<'a>
     pub resources : &'a mut DataMultistore
 }
 
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, Eq, PartialEq)]
 pub struct Color
 {
     pub r : u8,
@@ -404,6 +415,27 @@ impl Sheet
         }
     }
 
+    pub fn new(width : u32, height : u32, palette : Option<Vec<Color>>) -> Sheet
+    {
+        let length = match palette
+        {
+            None => { (width * height * 4) as usize }
+            Some(_) => { (width * height) as usize }
+        };
+
+        Sheet
+        {
+            width,
+            height,
+            palette,
+            image_raw : vec![0; length],
+            metadata : SheetMetadata::new(),
+            texture_dirty : true,
+            image_file_dirty : false,
+            generation : 1
+        }
+    }
+
     pub fn width(&self) -> u32 { self.width }
     pub fn height(&self) -> u32 { self.height }
 
@@ -478,6 +510,11 @@ impl Sheet
         self.metadata.slices.get_id(name)
     }
 
+    pub fn add_slice(&mut self, slice : SpriteSlice)
+    {
+        self.insert_slice(self.metadata.slices.len(), slice);
+    }
+
     pub fn insert_slice(&mut self, pos : usize, slice : SpriteSlice) -> bool
     {
         let res = self.metadata.slices.insert_item(pos, slice);
@@ -514,6 +551,12 @@ impl Sheet
         res
     }
 
+    pub fn clear_slices(&mut self)
+    {
+        self.metadata.slices.clear();
+        self.next_generation();
+    }
+
     pub fn slice_count(&self) -> usize
     {
         self.metadata.slices.len()
@@ -600,6 +643,28 @@ impl Sheet
     }
 }
 
+impl Clone for Sheet
+{
+    fn clone(&self) -> Self
+    {
+        let mut sheet = Sheet
+        {
+            width : self.width,
+            height : self.height,
+            palette : self.palette.clone(),
+            image_raw : self.image_raw.clone(),
+            metadata : self.metadata.clone(),
+            texture_dirty : true,
+            image_file_dirty : self.image_file_dirty,
+            generation : self.generation
+        };
+
+        sheet.next_generation();
+
+        sheet
+    }
+}
+
 impl DataObject for Sheet
 {
     fn folder_name() -> &'static str where Self : Sized
@@ -611,8 +676,41 @@ impl DataObject for Sheet
     {
         pathname.ends_with(".png")
     }
-    fn from_io(reader : Box<dyn ReadSeek>, full_pathname : &str, source : &mut Box<dyn Source>) -> Result<Self, DataError> where Self : Sized
+    fn from_package_source(source : &mut Box<dyn Source>, package_name : &str, pathname : &str) -> Result<Self, DataError> where Self : Sized
     {
+        let mut metadata = SheetMetadata::new();
+
+        let mut new_path = String::from(pathname);
+        new_path.push_str(".json");
+
+        if let Ok(mut json_read) = source.read_file(package_name, &new_path)
+        {
+            let mut json_text = String::new();
+            let metadata_result = json_read.read_to_string(&mut json_text);
+
+            match metadata_result
+            {
+                Err(error) =>
+                    {
+                        return Err(DataError::IoError(error));
+                    },
+                _ => {}
+            }
+
+            match serde_json::from_str(&json_text)
+            {
+                Ok(loaded_metadata) =>
+                {
+                    metadata = loaded_metadata;
+                },
+                Err(error) =>
+                {
+                    return Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error)));
+                }
+            }
+        }
+
+        let reader = source.read_file(package_name, pathname)?;
         let mut decoder = png::Decoder::new(reader);
 
         decoder.set_transformations(png::Transformations::IDENTITY);
@@ -644,43 +742,11 @@ impl DataObject for Sheet
                 {
                     Err(error) =>
                     {
-                        return Err(DataError::BadData(format!("{}: {}", full_pathname, error)));
+                        return Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error)));
                     },
                     _ => {}
                 }
 
-                let mut metadata = SheetMetadata::new();
-
-                let mut new_path = String::from(full_pathname);
-                new_path.push_str(".json");
-
-                if let Ok(mut json_read) = source.read_file(&new_path)
-                {
-                    let mut json_text = String::new();
-                    let metadata_result = json_read.read_to_string(&mut json_text);
-
-                    match metadata_result
-                    {
-                        Err(error) =>
-                        {
-                            return Err(DataError::IoError(error));
-                        },
-                        _ => {}
-                    }
-
-                    match serde_json::from_str(&json_text)
-                    {
-                        Ok(loaded_metadata) =>
-                        {
-                            metadata = loaded_metadata;
-                        },
-                        Err(error) =>
-                        {
-                            return Err(DataError::BadData(format!("{}: {}", full_pathname, error)));
-                        }
-                    }
-                }
-
                 return Ok(Sheet
                 {
                     width : info.width as u32,
@@ -699,16 +765,16 @@ impl DataObject for Sheet
             },
             Err(error) =>
             {
-                return Err(DataError::BadData(format!("{}: {}", full_pathname, error)));
+                return Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error)));
             }
-        }
+        };
     }
 
-    fn write(&mut self, full_pathname : &str, source : &mut Box<dyn Source>) -> Result<(), DataError>
+    fn write(&mut self, package_name : &str, pathname : &str, source : &mut Box<dyn Source>) -> Result<(), DataError>
     {
         if self.image_file_dirty
         {
-            let writer = try_or_else!(source.write_file(full_pathname),
+            let writer = try_or_else!(source.write_file(package_name, pathname),
                 |error| Err(DataError::PackageSourceError(error)));
 
             let mut encoder = png::Encoder::new(writer, self.width, self.height);
@@ -736,7 +802,7 @@ impl DataObject for Sheet
                     {
                         Ok(_) =>
                         {
-                            info!("Wrote sheet PNG to {}", full_pathname);
+                            info!("Wrote sheet PNG to {}/{}", package_name, pathname);
                         },
                         Err(png::EncodingError::IoError(io_error)) =>
                         {
@@ -744,7 +810,7 @@ impl DataObject for Sheet
                         },
                         Err(error) =>
                         {
-                            return Err(DataError::BadData(format!("{}: {}", full_pathname, error)));
+                            return Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error)));
                         }
                     }
                 },
@@ -754,7 +820,7 @@ impl DataObject for Sheet
                 },
                 Err(error) =>
                 {
-                    return Err(DataError::BadData(format!("{}: {}", full_pathname, error)));
+                    return Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error)));
                 }
             }
         }
@@ -763,14 +829,14 @@ impl DataObject for Sheet
             info!("Sheet image unchanged, not writing PNG");
         }
 
-        let mut new_path = String::from(full_pathname);
+        let mut new_path = String::from(pathname);
         new_path.push_str(".json");
 
-        let mut writer = try_or_else!(source.write_file(&new_path),
+        let mut writer = try_or_else!(source.write_file(package_name, &new_path),
                 |error| Err(DataError::PackageSourceError(error)));
 
         let json_string = try_or_else!(serde_json::to_string(&self.metadata),
-            |error| Err(DataError::BadData(format!("{}: {}", full_pathname, error))));
+            |error| Err(DataError::BadData(format!("{}/{}: {}", package_name, pathname, error))));
 
         match writer.write_all(json_string.as_bytes())
         {
@@ -797,7 +863,7 @@ impl DataObject for Sheet
     }
 }
 
-#[derive(Serialize, Deserialize, Debug)]
+#[derive(Clone, Serialize, Deserialize, Debug)]
 pub struct SheetMetadata
 {
     #[serde(rename = "txfl")]
@@ -838,8 +904,9 @@ impl DataObject for Shader
     {
         pathname.ends_with(".glsl")
     }
-    fn from_io(mut reader: Box<dyn ReadSeek>, _full_pathname : &str, _source : &mut Box<dyn Source>) -> Result<Self, DataError> where Self : Sized
+    fn from_package_source(source : &mut Box<dyn Source>, package_name : &str, pathname : &str) -> Result<Self, DataError> where Self : Sized
     {
+        let mut reader = source.read_file(package_name, pathname)?;
         let mut program_source = String::new();
         match reader.read_to_string(&mut program_source)
         {
@@ -1067,6 +1134,8 @@ impl Renderer
         let mut fullscreen = false;
         let mut kiosk_mode = false;
         let mut one_to_one_viewport = false;
+        let mut base_width = 100.0;
+        let mut base_height = 100.0;
         
         if let Some(mut new_info) = renderer_new_info
         {
@@ -1090,6 +1159,8 @@ impl Renderer
                         fullscreen = new_info.fullscreen;
                         kiosk_mode = new_info.kiosk_mode;
                         one_to_one_viewport = new_info.viewport_mode == ViewportMode::OneToOne;
+                        base_width = new_info.width as f32;
+                        base_height = new_info.height as f32;
                         
                         info!("Using renderer backend \"{}\"", backend.name());
                         
@@ -1103,13 +1174,17 @@ impl Renderer
                 }
             }
 
+            #[allow(unused_variables)]
             if let Some(window) = &window
             {
-                let (mut icon_raw, icon_width, icon_height) = png_to_raw(Cursor::new(include_bytes!("../res/icon.png").to_vec()))?;
-                let surface = try_or_else!(Surface::from_data(&mut icon_raw, icon_width, icon_height, icon_width * 4, PixelFormatEnum::ABGR8888),
-                    |error| Err(RendererError::LoadResourceFailed(format!("Failed to create raw surface: {}", error))));
+                #[cfg(not(windows))]
+                {
+                    let (mut icon_raw, icon_width, icon_height) = png_to_raw(Cursor::new(include_bytes!("../res/icon.png").to_vec()))?;
+                    let surface = try_or_else!(Surface::from_data(&mut icon_raw, icon_width, icon_height, icon_width * 4, PixelFormatEnum::ABGR8888),
+                        |error| Err(RendererError::LoadResourceFailed(format!("Failed to create raw surface: {}", error))));
 
-                window.borrow_mut().set_icon(surface);
+                    window.borrow_mut().set_icon(surface);
+                }
             }
             else
             {
@@ -1122,12 +1197,12 @@ impl Renderer
             control,
             window,
             video_subsystem,
-            base_width : 100.0,
-            base_height : 100.0,
+            base_width,
+            base_height,
             one_to_one_viewport,
             kiosk_mode
         };
-        
+
         renderer.recalculate_viewport();
         renderer.set_fullscreen(fullscreen);
         renderer.set_visible(true);
@@ -1166,6 +1241,17 @@ impl Renderer
         
         (0, 0)
     }
+
+    pub fn window_scissor(&self) -> Rect<u32>
+    {
+        let (window_w, window_h) = self.window_size();
+
+        // Forcing Independent viewport mode because we are getting window viewport, not drawing viewport
+        let info = crate::renderer::recalculate_viewport_common(self.base_width, self.base_height,
+                                                                window_w, window_h, ViewportMode::Independent);
+
+        info.scissor.unwrap_or_default()
+    }
     
     pub fn draw(&mut self, scene : &mut Box<dyn BaseScene>, resources : &mut DataMultistore, interpolation : f32)
     {
@@ -1318,6 +1404,26 @@ impl Renderer
         }
     }
 
+    pub fn get_clipboard(&self) -> Option<String>
+    {
+        if let Some(video_subsystem) = &self.video_subsystem
+        {
+            return video_subsystem.clipboard().clipboard_text().ok();
+        }
+
+        None
+    }
+
+    pub fn set_clipboard(&self, text : &str) -> bool
+    {
+        if let Some(video_subsystem) = &self.video_subsystem
+        {
+            return video_subsystem.clipboard().set_clipboard_text(text).is_ok();
+        }
+
+        false
+    }
+
     pub(crate) fn current_refresh_rate(&self) -> i32
     {
         if let Some(window) = &self.window
