@@ -294,6 +294,18 @@ impl DataPreparer<Sheet, GpuSheetHandle> for GpuSheetPreparer
     }
 }
 
+#[allow(unused_mut)]
+fn modify_shader_header(mut shader_text : &str) -> String
+{
+    #[cfg(not(target_os = "emscripten"))]
+    {
+        shader_text.to_string()
+    }
+    #[cfg(target_os = "emscripten")]
+    {
+        shader_text.replace("#version 140", "#version 300 es\nprecision mediump float;")
+    }
+}
 
 struct GpuShaderHandle
 {
@@ -310,8 +322,8 @@ impl DataPreparer<Shader, GpuShaderHandle> for GpuShaderPreparer
     fn prepare(&mut self, data : &mut Shader, _id : DataId) -> GpuShaderHandle
     {
         let program = make_program(&self.display.borrow(),
-            include_str!("shader/vert_standard_140.glsl"),
-            &data.program_source).expect("Shader compilation failed");
+            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
+            &modify_shader_header(&data.program_source)).expect("Shader compilation failed");
         
         GpuShaderHandle
         {
@@ -514,7 +526,7 @@ impl GlDrawControl
         
         self.current_index = 0;
 
-        self.instance_buffer_flip = (self.instance_buffer_flip + 2) % self.instance_buffer.len();
+        self.flip_instance_buffers(2);
     }
 
     fn flush_drawing(&mut self)
@@ -523,6 +535,11 @@ impl GlDrawControl
 
         self.display.borrow().get_context().flush();
     }
+
+    fn flip_instance_buffers(&mut self, amount : usize)
+    {
+        self.instance_buffer_flip = (self.instance_buffer_flip + amount) % self.instance_buffer.len();
+    }
     
     fn update_pixel_upscale(&mut self)
     {
@@ -552,9 +569,17 @@ impl DrawControl for GlDrawControl
         renderer_new_info.height * renderer_new_info.default_zoom);
         
         window_builder.hidden().position_centered().resizable().allow_highdpi();
-        
-        video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::Core);
-        video_subsystem.gl_attr().set_context_version(3, 2);
+
+        #[cfg(not(target_os = "emscripten"))]
+        {
+            video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::Core);
+            video_subsystem.gl_attr().set_context_version(3, 2);
+        }
+        #[cfg(target_os = "emscripten")]
+        {
+            video_subsystem.gl_attr().set_context_profile(sdl2::video::GLProfile::GLES);
+            video_subsystem.gl_attr().set_context_version(3, 0);
+        }
         
         let display_raw = try_or_else!(window_builder.build_glium(),
             |error| Err(RendererError::RendererInitFailed(format!("Could not create game window: {}", error))));
@@ -595,12 +620,12 @@ impl DrawControl for GlDrawControl
         }
         
         let shader_background = try_or_else!(make_program(&display_raw,
-            include_str!("shader/vert_standard_140.glsl"),
-            include_str!("shader/frag_solidcolor_140.glsl")),
+            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
+            &modify_shader_header(include_str!("shader/frag_solidcolor_140.glsl"))),
             |error| Err(RendererError::RendererInitFailed(format!("Could not load shader: {}", error))));
         let shader_sprite = try_or_else!(make_program(&display_raw,
-            include_str!("shader/vert_standard_140.glsl"),
-            include_str!("shader/frag_sprite_140.glsl")),
+            &modify_shader_header(include_str!("shader/vert_standard_140.glsl")),
+            &modify_shader_header(include_str!("shader/frag_sprite_140.glsl"))),
             |error| Err(RendererError::RendererInitFailed(format!("Could not load shader: {}", error))));
         
         let shader_list = vec!(shader_background, shader_sprite);
@@ -909,7 +934,7 @@ impl DrawControl for GlDrawControl
         }
         
         self.last_view_transform = view_transform.clone();
-        self.instance_buffer_flip = (self.instance_buffer_flip + 1) % self.instance_buffer.len();
+        self.flip_instance_buffers(1);
     }
 
     fn end_drawing(&mut self)
@@ -931,10 +956,10 @@ impl DrawControl for GlDrawControl
                     let mut transform = DrawTransform::from_array(&UNBORKED_UPSCALE_TRANSFORM);
                     
                     let params = glium::DrawParameters::default();
-                    let instance = upscale_instance_from_transform(&transform);
-                    
+
+                    self.instance_buffer_src[0] = upscale_instance_from_transform(&transform);
                     self.instance_buffer[self.instance_buffer_flip].invalidate();
-                    self.instance_buffer[self.instance_buffer_flip].map_write().set(0, instance);
+                    self.instance_buffer[self.instance_buffer_flip].write(&self.instance_buffer_src[..]);
                     
                     self.update_pixel_upscale();
                     
@@ -966,12 +991,13 @@ impl DrawControl for GlDrawControl
                     let (x, y) = self.viewport_offset;
                     transform.scale(1.0 / self.scale_width, 1.0 / self.scale_height);
                     transform.translate(x, y);
-                    let instance = upscale_instance_from_transform(&transform);
-                    
-                    self.instance_buffer[self.instance_buffer_flip].map_write().set(1, instance);
+
+                    self.instance_buffer_src[0] = upscale_instance_from_transform(&transform);
+                    self.instance_buffer[self.instance_buffer_flip].invalidate();
+                    self.instance_buffer[self.instance_buffer_flip].write(&self.instance_buffer_src[..]);
                     
                     self.frame.as_mut().unwrap().draw((&self.quad_buffer,
-                        self.instance_buffer[self.instance_buffer_flip].slice(1..2)
+                        self.instance_buffer[self.instance_buffer_flip].slice(0..1)
                             .expect("Failed to prepare instance buffer").per_instance()
                             .expect("Failed to prepare instance buffer")),
                         &self.index_buffer, &self.shader_list[RendererOp::Sprite as usize],
