@@ -1,4 +1,4 @@
-use crate::gameloop::{GameInfo, GameControl};
+use crate::gameloop::{GameInfo, GameControl, GameControlOptions};
 use crate::renderer::{DrawTransform, DrawBackgroundColorInfo};
 use crate::scene::{BaseScene, SimpleScene, NullSceneType, Entity, Component, ThinkerSystem, DrawerSystem, ThinkerArgs, DrawerArgs};
 use rand::seq::SliceRandom;
@@ -246,7 +246,7 @@ fn scene_empty()
     assert_eq!(scene.entity_count(), 0);
     
     // Check empty state doesn't break when used for weird stuff
-    scene.think(&mut GameControl::new(&GameInfo::default(), None).unwrap());
+    scene.think(&mut GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap());
     
     let bad_handle = Entity::new(0, 0);
     
@@ -311,7 +311,7 @@ fn scene_null_entity()
 fn scene_basic_system()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
     
     let entity = scene.add_entity();
     assert_eq!(scene.add_component(&entity, FooComponent::new()), true);
@@ -347,7 +347,7 @@ fn scene_basic_system()
 fn scene_multi_component_system()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -399,7 +399,7 @@ fn scene_multi_component_system()
 fn scene_remove_entity_component()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -530,7 +530,7 @@ fn scene_remove_random_ordered()
     for _ in 0..100
     {
         let mut scene = SimpleScene::new();
-        let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+        let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
         let mut entities = Vec::new();
         let rng = &mut rand::thread_rng();
         
@@ -599,7 +599,7 @@ fn scene_remove_random_ordered()
 fn scene_component_registration()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
 
     scene.add_thinker_system(Box::new(FooSystem::new()));
     scene.add_thinker_system(Box::new(BarAdderSystem::new()));
@@ -627,7 +627,7 @@ fn scene_component_registration()
 fn scene_immutable_draw()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -651,7 +651,7 @@ fn scene_immutable_draw()
 fn scene_mutable_draw_should_panic()
 {
     let mut scene = SimpleScene::new();
-    let mut gameloop = GameControl::new(&GameInfo::default(), None).unwrap();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
     
     let entity1 = scene.add_entity();
     let entity2 = scene.add_entity();
@@ -671,3 +671,204 @@ fn scene_mutable_draw_should_panic()
 
     std::mem::swap(&mut res, gameloop.res_mut());
 }
+
+#[derive(Debug)]
+struct DeferredSystem
+{
+    runs : u32,
+    foo_entity : Entity,
+    blank_entity : Entity,
+    foobar_entity : Entity,
+    foobar2_entity : Entity,
+    ghost_entity : Entity,
+    spaced_entity : Entity,
+}
+
+impl DeferredSystem
+{
+    fn new() -> DeferredSystem
+    {
+        DeferredSystem
+        {
+            runs : 0,
+            foo_entity : Entity::null(),
+            blank_entity : Entity::null(),
+            foobar_entity : Entity::null(),
+            foobar2_entity : Entity::null(),
+            ghost_entity : Entity::null(),
+            spaced_entity : Entity::null(),
+        }
+    }
+}
+
+impl ThinkerSystem<NullSceneType> for DeferredSystem
+{
+    fn think(&mut self, args : ThinkerArgs<NullSceneType>)
+    {
+        match self.runs
+        {
+            0 =>
+            {
+                let foo_store = args.components.store_mut::<FooComponent>();
+                let bar_store = args.components.store_mut::<BarComponent>();
+
+                assert_eq!(foo_store.count(), 0);
+                assert_eq!(bar_store.count(), 0);
+
+                self.foo_entity = args.scene.add_entity_later();
+                args.scene.add_component_later(&self.foo_entity, FooComponent::new());
+                self.blank_entity = args.scene.add_entity_later(); // Don't do this in real programs :)
+                self.foobar_entity = args.scene.add_entity_later();
+                args.scene.add_component_later(&self.foobar_entity, FooComponent::new());
+                args.scene.add_component_later(&self.foobar_entity, BarComponent { runs : 1 });
+                // Should fail later
+                args.scene.add_component_later(&self.foobar_entity, BarComponent { runs : 2 });
+                self.foobar2_entity = args.scene.add_entity_later();
+                args.scene.add_component_later(&self.foobar2_entity, FooComponent::new());
+                args.scene.add_component_later(&self.foobar2_entity, BarComponent { runs : 3 });
+
+                assert_eq!(foo_store.count(), 0);
+                assert_eq!(bar_store.count(), 0);
+                assert!(foo_store.get_entity(&self.foo_entity).is_none());
+                assert!(foo_store.get_entity(&self.foobar_entity).is_none());
+                assert!(bar_store.get_entity(&self.foobar_entity).is_none());
+                assert!(foo_store.get_entity(&self.foobar2_entity).is_none());
+                assert!(bar_store.get_entity(&self.foobar2_entity).is_none());
+            }
+            1 =>
+            {
+                let foo_store = args.components.store_mut::<FooComponent>();
+                let bar_store = args.components.store_mut::<BarComponent>();
+
+                assert_eq!(foo_store.count(), 3);
+                assert_eq!(bar_store.count(), 2);
+                assert!(foo_store.get_entity(&self.foo_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar2_entity).is_some());
+                assert!(bar_store.get_entity(&self.foo_entity).is_none());
+                assert!(bar_store.get_entity(&self.foobar_entity).is_some());
+                assert!(bar_store.get_entity(&self.foobar2_entity).is_some());
+
+                let bar = bar_store.get_entity(&self.foobar_entity).unwrap();
+
+                // Check that the first one, not the second one, was added
+                assert_eq!(bar.runs, 1);
+
+                args.scene.remove_component_later::<FooComponent>(&self.foobar_entity);
+                args.scene.remove_entity_later(&self.foobar2_entity);
+
+                assert_eq!(foo_store.count(), 3);
+                assert_eq!(bar_store.count(), 2);
+                assert!(foo_store.get_entity(&self.foo_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar2_entity).is_some());
+                assert!(bar_store.get_entity(&self.foo_entity).is_none());
+                assert!(bar_store.get_entity(&self.foobar_entity).is_some());
+                assert!(bar_store.get_entity(&self.foobar2_entity).is_some());
+            }
+            2 =>
+            {
+                let foo_store = args.components.store_mut::<FooComponent>();
+                let bar_store = args.components.store_mut::<BarComponent>();
+
+                assert_eq!(foo_store.count(), 1);
+                assert_eq!(bar_store.count(), 1);
+                assert!(foo_store.get_entity(&self.foo_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar_entity).is_none());
+                assert!(foo_store.get_entity(&self.foobar2_entity).is_none());
+                assert!(bar_store.get_entity(&self.foo_entity).is_none());
+                assert!(bar_store.get_entity(&self.foobar_entity).is_some());
+                assert!(bar_store.get_entity(&self.foobar2_entity).is_none());
+
+                args.scene.add_component_later(&self.foobar_entity, FooComponent { runs : 4 });
+                self.foobar2_entity = args.scene.add_entity_later();
+                args.scene.add_component_later(&self.foobar2_entity, FooComponent::new());
+                args.scene.add_component_later(&self.foobar2_entity, BarComponent { runs : 5 });
+            }
+            3 =>
+            {
+                let foo_store = args.components.store_mut::<FooComponent>();
+                let bar_store = args.components.store_mut::<BarComponent>();
+
+                assert_eq!(foo_store.count(), 3);
+                assert_eq!(bar_store.count(), 2);
+                assert!(foo_store.get_entity(&self.foo_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar_entity).is_some());
+                assert!(foo_store.get_entity(&self.foobar2_entity).is_some());
+                assert!(bar_store.get_entity(&self.foo_entity).is_none());
+                assert!(bar_store.get_entity(&self.foobar_entity).is_some());
+                assert!(bar_store.get_entity(&self.foobar2_entity).is_some());
+
+                let foo = foo_store.get_entity(&self.foobar_entity).unwrap();
+
+                assert_eq!(foo.runs, 4);
+
+                // Now we try some weird stuff and make sure it still has consistent results
+
+                self.ghost_entity = args.scene.add_entity_later();
+                args.scene.add_component_later(&self.ghost_entity, FooComponent::new());
+                args.scene.remove_entity_later(&self.ghost_entity);
+                args.scene.add_component_later(&self.ghost_entity, BarComponent::new());
+            }
+            4 =>
+            {
+                let foo_store = args.components.store_mut::<FooComponent>();
+                let bar_store = args.components.store_mut::<BarComponent>();
+
+                assert_eq!(foo_store.count(), 3);
+                assert_eq!(bar_store.count(), 2);
+
+                assert!(foo_store.get_entity(&self.ghost_entity).is_none());
+                assert!(bar_store.get_entity(&self.ghost_entity).is_none());
+
+                args.scene.remove_entity_later(&self.ghost_entity);
+                self.ghost_entity = args.scene.add_entity_later();
+
+                for i in 0..100
+                {
+                    args.scene.remove_entity_later(&self.spaced_entity);
+                    self.spaced_entity = args.scene.add_entity_later();
+                    args.scene.add_component_later(&self.spaced_entity, BarComponent { runs : i });
+                }
+            }
+            5 =>
+            {
+                let foo_store = args.components.store_mut::<FooComponent>();
+                let bar_store = args.components.store_mut::<BarComponent>();
+
+                assert_eq!(foo_store.count(), 3);
+                assert_eq!(bar_store.count(), 3);
+                assert!(bar_store.get_entity(&self.spaced_entity).is_some());
+
+                let bar = bar_store.get_entity(&self.spaced_entity).unwrap();
+
+                assert_eq!(bar.runs, 99);
+            }
+            _ => {}
+        }
+
+        self.runs += 1;
+    }
+}
+
+#[test]
+fn scene_deferred_add_remove()
+{
+    let mut scene = SimpleScene::new();
+    let mut gameloop = GameControl::new(&GameInfo::default(), GameControlOptions::headless()).unwrap();
+
+    scene.add_thinker_system(Box::new(DeferredSystem::new()));
+
+    assert_eq!(scene.entity_count(), 0);
+    scene.think(&mut gameloop);
+    assert_eq!(scene.entity_count(), 4);
+    scene.think(&mut gameloop);
+    assert_eq!(scene.entity_count(), 3);
+    scene.think(&mut gameloop);
+    assert_eq!(scene.entity_count(), 4);
+    scene.think(&mut gameloop);
+    assert_eq!(scene.entity_count(), 4);
+    scene.think(&mut gameloop);
+    assert_eq!(scene.entity_count(), 6);
+    scene.think(&mut gameloop);
+}
